/*
 * Copyright 2011 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.robokind.integration.animation_motion;

import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.jms.Connection;
import javax.jms.JMSException;
import org.jflux.api.core.config.Configuration;
import org.jflux.impl.transport.jms.ConnectionConfigUtils;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.robokind.api.animation.lifecycle.AnimationPlayerClientLifecycle;
import org.robokind.api.animation.messaging.RemoteAnimationPlayerClient;
import org.robokind.api.animation.player.AnimationPlayer;
import org.robokind.api.animation.protocol.AnimationEvent;
import org.robokind.api.animation.protocol.AnimationEvent.AnimationEventFactory;
import org.robokind.api.animation.utils.AnimationEditListener;
import org.robokind.api.animation.utils.PositionAdder;
import org.robokind.api.common.lifecycle.ManagedService;
import org.robokind.api.common.lifecycle.utils.SimpleLifecycle;
import org.robokind.api.common.osgi.OSGiUtils;
import org.robokind.api.common.osgi.lifecycle.OSGiComponent;
import org.robokind.api.motion.Robot;
import org.robokind.api.motion.blending.FrameSource;
import org.robokind.api.motion.messaging.RemoteRobot;
import org.robokind.api.motion.utils.RobotUtils;
import org.robokind.avrogen.animation.AnimationRecord;
import org.robokind.impl.animation.messaging.PortableAnimationEvent;
import org.robokind.impl.messaging.config.RKMessagingConfigUtils;
import org.robokind.impl.messaging.lifecycle.JMSAvroMessageSenderLifecycle;
import org.robokind.impl.messaging.utils.ConnectionManager;
import org.robokind.impl.messaging.utils.ConnectionUtils;
import org.robokind.integration.animation_motion.lifecycle.RobotChannelsLifecycle;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public class AnimationMotionUtils {
    private final static Logger theLogger = Logger.getLogger(AnimationMotionUtils.class.getName());
    
    public final static long DEFAULT_BLENDER_INTERVAL = 40L;
    public final static int DEFAULT_MAX_RAMP_TIME_MILLISEC = 800;
    
    public static ServiceRegistration registerRobotAnimationPlayer(
            BundleContext context, Robot.Id robotId){
        Boolean exists = OSGiUtils.serviceExists(
                context, AnimationPlayer.class.getName(),
                RobotUtils.getRobotFilter(robotId));
        if(exists == null || exists){
            return null;
        }
        return registerAnimationPlayer(context, robotId);
    }
    
    public static ServiceRegistration[] registerAnimationEditingComponents(
            BundleContext context, Robot robot){
//        String[] classes = new String[]{
//                ChannelsParameterSource.class.getName(), 
//                AnimationEditListener.class.getName(), 
//                PositionAdder.class.getName()};
//        Boolean exists = OSGiUtils.serviceExists(context, classes, null);
//        if(exists == null || exists){
//            return null;
//        }
        Robot.Id robotId = robot.getRobotId();
        //boolean player = OSGiUtils.serviceExists(
        //        context, AnimationPlayer.class, null);
        int n = 4;//player ? 3 : 4;
        ServiceRegistration[] regs = new ServiceRegistration[n];
        regs[0] = registerChannelRobotParams(context, robotId);
        regs[1] = registerAnimationEditListener(context, robot);
        regs[2] = registerJointSnapshotAdder(context, robotId);
        if(!OSGiUtils.serviceExists(context, RemoteAnimationPlayerClient.class, null)){
            if(!(robot instanceof RemoteRobot)){
                regs[3] = registerAnimationPlayer(context, robotId);
            }else{
                registerRemoteAnimPlayer(context, "robotReceiverConnectionConfig");
            }
        }
        return regs;
    }
    
    private static ServiceRegistration registerAnimationPlayer(
            BundleContext context, Robot.Id robotId){
        AnimationPlayer player = new RampedAnimationPlayer(
                context, robotId, DEFAULT_MAX_RAMP_TIME_MILLISEC);
        Properties props = new Properties();
        String playerId = player.getAnimationPlayerId();
        props.put(Robot.PROP_ID, robotId.toString());
        ServiceRegistration reg = OSGiUtils.registerUniqueService(context, 
                AnimationPlayer.class.getName(), 
                AnimationPlayer.PROP_PLAYER_ID, 
                playerId, player, props);
        theLogger.log(Level.INFO, 
                "OSGiAnimationPlayer Service Registered Successfully.");
        return reg;
    }
    
    private static synchronized void registerRemoteAnimPlayer(BundleContext context, String connectionConfigId){
        ServiceReference[] confRefs;
        try{
            confRefs = context.getAllServiceReferences(
                        Configuration.class.getName(), 
                        OSGiUtils.createFilter(
                                RKMessagingConfigUtils.JMS_CONNECTION_CONFIG, 
                                connectionConfigId));
        }catch(InvalidSyntaxException ex){
            theLogger.log(Level.WARNING, "Unable to find connection information to create remote animation player");
            return;
        }
        if(confRefs == null || confRefs.length == 0){
            theLogger.log(Level.WARNING, "Unable to find connection information to create remote animation player");
            return;
        }
        ServiceReference conConfRef = confRefs[0];
        Configuration<String> conConf = 
                OSGiUtils.getService(Configuration.class, context, conConfRef);
        String ip = conConf.getPropertyValue(String.class, ConnectionConfigUtils.CONF_BROKER_IP);
        Connection con = ConnectionManager.createConnection(
                "admin", "admin", "client1", "test", 
                "tcp://" + ip + ":5672");
        try{
            con.start();
        }catch(JMSException ex){
            theLogger.log(Level.WARNING, "Unable to connect to {0}", ip);
            return;
        }
        ManagedService myConnectionService = new OSGiComponent(context, new SimpleLifecycle(con, Connection.class));
        myConnectionService.start();
        ConnectionUtils.ensureSession(context, 
                "remoteAnimConnection", con, null);
        ConnectionUtils.ensureDestinations(context, 
                "remoteAnimationRequest", "animationRequest", ConnectionUtils.TOPIC, null);
        JMSAvroMessageSenderLifecycle senderLife = 
                new JMSAvroMessageSenderLifecycle(
                        new PortableAnimationEvent.MessageRecordAdapter(), 
                        AnimationEvent.class, AnimationRecord.class, 
                        "remoteAnimSender", "remoteAnimConnection", 
                        "remoteAnimationRequest");
        registerEventFactory(context);
        ManagedService mySenderService = new OSGiComponent(context, senderLife);
        mySenderService.start();
        AnimationPlayerClientLifecycle myLifecycle = 
                new AnimationPlayerClientLifecycle(
                        "remotePlayer", "remotePlayer", "remoteAnimSender");
        ManagedService myPlayerService = new OSGiComponent(context, myLifecycle);
        myPlayerService.start();
    }
    
    private static void registerEventFactory(BundleContext context){
        if(OSGiUtils.serviceExists(context, AnimationEvent.AnimationEventFactory.class, null)){
            return;
        }
        new OSGiComponent(context, 
                new SimpleLifecycle(
                        new PortableAnimationEvent.Factory(), 
                        AnimationEventFactory.class)
                ).start();
    }
    
    private static ServiceRegistration registerChannelRobotParams(
            BundleContext context, Robot.Id robotId){
        RobotChannelsLifecycle lifecycle = new RobotChannelsLifecycle(robotId);
        OSGiComponent comp = new OSGiComponent(context, lifecycle);
        comp.start();
        return null;
    }
    
    private static ServiceRegistration registerAnimationEditListener(
            BundleContext context, Robot robot){
        AnimationEditFrameSource aefs = 
                new AnimationEditFrameSource(robot);
        String[] classes = {
            AnimationEditListener.class.getName(), 
            FrameSource.class.getName()};
        Dictionary<String, Object> props = new Hashtable<String, Object>();
        props.put(Robot.PROP_ID, robot.getRobotId().toString());
        
        ServiceRegistration reg = context.registerService(classes, aefs, props);
        theLogger.log(Level.INFO, 
                "AnimationEditFrameSource Service Registered Successfully.");
        return reg;
    }
    
    private static ServiceRegistration registerJointSnapshotAdder(
            BundleContext context, Robot.Id robotId){
        JointSnapshot js = new JointSnapshot(context, robotId);
        Dictionary<String, Object> props = new Hashtable<String, Object>();
        props.put(Robot.PROP_ID, robotId.toString());
        ServiceRegistration reg = context.registerService(
                PositionAdder.class.getName(), js, props);
        theLogger.log(Level.INFO, 
                "JointSnapshot Service Registered Successfully.");
        return reg;
    }
}
